--	Written by Murray Steele for the Edinburgh University Haskell programming competition
--	Fractal Terrain Generator


-----------------
--	TODO:
--	>Make main function
----------------
import Data.List
import System.IO

data Terrain 		= Node Double Terrain Terrain Terrain Terrain | Point Double

type Heightmap		= [[Double]]
type Positionmap	= [[(Double, Double, Double)]]
type Vector			= (Double, Double, Double)
type RVector		= (Int, Int)
type Matrix			= [[Double]]
type Boolmap		= [[Bool]]
type Face			= ((Vector, Vector, Vector), Double)
type RFVector		= (Int, Int, Int)
type RFace			= ((RFVector, RFVector, RFVector), Double)
type RFPixel		= ((Int, Int), RGB)
type RGB			= (Int, Int, Int)

main = do
	putStr "\nWelcome to terrain-generator!\n\nThis uses a tree based diamond square algorithm to create fractal landscapes!\n\n"
	putStr "You will be asked to enter parameters for the algorithm.\nPlease keep in mind that large values of some parameters\nwill slow down the terrain generation. Sometimes you\nmay need to wait a few minutes for an image to be outputted.\n\nExample parameters and advice can be found in the README.txt\nthat was included with this.\nSeveral example generated images were also included.\n\n"
	putStr "The image format chosen was PPM for its simplicity.\nIf you cannot view PPM images then there are many\nutilities availiable online that can be used to convert\nit to a format you can view.\n\n"
	putStrLn "Enter depth: "
	depth <- getLine
	putStrLn "Enter environments: "
	envs <- getLine
	putStrLn "Enter scale: "
	scale <- getLine
	generateWireEnvironments (fromIntegral (read depth :: Int)) (read envs :: [[(Double, Double, Double, Double, Double)]]) (read scale :: Double) (pi/4, 0, pi/4)
	putStr "WIRE ENVIRONMENT COMPLETED\n"
	generateFacedEnvironments (fromIntegral (read depth :: Int)) (read envs :: [[(Double, Double, Double, Double, Double)]]) (read scale :: Double) (pi/4, 0, pi/4)
	putStr "FACED ENVIRONMENT COMPLETED\n"

generateFacedEnvironments :: Double -> [[(Double, Double, Double, Double, Double)]] -> Double -> (Double, Double, Double) -> IO()
generateFacedEnvironments depth envs scale rotation = (writePPM "faced-environment.ppm" . rfpixelsToPPM . rfacesToFPixels (0, 0, 1) . fsToRFs . getFaces . rotatePM rotation . scalePM scale . toPositionmap . stitchHeightmapMap . map (map (makeHeightmap depth))) envs

generateWireEnvironments :: Double -> [[(Double, Double, Double, Double, Double)]] -> Double -> (Double, Double, Double) -> IO()
generateWireEnvironments depth envs scale rotation =  (writePPM "wire-environment.ppm" . pixelsToPPM . getPixels . vsToRVs . getVectors . rotatePM rotation . scalePM scale . toPositionmap . stitchHeightmapMap . map (map (makeHeightmap depth))) envs

makeHeightmap :: Double -> (Double, Double, Double, Double, Double) -> Heightmap
makeHeightmap depth (a, b, c, d, e) = toHeightmap (makeTerrain' depth a b c d) e

stitchHeightmapMap :: [[Heightmap]] -> Heightmap
stitchHeightmapMap = (onTopHeightmapColumn . map besideHeightmapRow)

onTopHeightmapColumn :: [Heightmap] -> Heightmap
onTopHeightmapColumn []		=  []
onTopHeightmapColumn (pm:pms)	= onTop pm (onTopHeightmapColumn pms)

besideHeightmapRow :: [Heightmap] -> Heightmap
besideHeightmapRow []			= []
besideHeightmapRow (pm:pms)	= beside pm (besideHeightmapRow pms)

makeTerrain :: Double -> Double -> Double -> Double -> Double -> Terrain
makeTerrain d tl tr bl br = (Node 1 (makeDepth d tl tl tr bl br) (makeDepth d tr tl tr bl br) (makeDepth d bl tl tr bl br) (makeDepth d br tl tr bl br))

makeTerrain' :: Double -> Double -> Double -> Double -> Double -> Terrain
makeTerrain' d tl tr bl br = (Node 1 (makeDepth' d tl tr tl br bl) (makeDepth' d tr tl tr bl br) (makeDepth' d bl br bl tr tl) (makeDepth' d br bl br tl tr))

makeDepth :: Double -> Double -> Double -> Double -> Double -> Double -> Terrain
makeDepth 0 h tl tr bl br = (Node h (Point tl) (Point tr) (Point bl) (Point br))
makeDepth d h tl tr bl br = (Node h (makeDepth (d - 1) tl tl tr bl br) (makeDepth (d - 1) tr tl tr bl br) (makeDepth (d - 1) bl tl tr bl br) (makeDepth (d - 1) br tl tr bl br))

makeDepth' :: Double -> Double -> Double -> Double -> Double -> Double -> Terrain
makeDepth' 0 h tl tr bl br = (Node h (Point tl) (Point tr) (Point bl) (Point br))
makeDepth' d h tl tr bl br = (Node h (makeDepth' (d - 1) tl tr tl br bl) (makeDepth' (d - 1) tr tl tr bl br) (makeDepth' (d - 1) bl br bl tr tl) (makeDepth' (d - 1) br bl br tl tr))

toHeightmap :: Terrain -> Double -> Heightmap
toHeightmap (Point h) e	= [[e * h]]
toHeightmap (Node h tl tr bl br) e = onTop (beside (scalarAdd c (toHeightmap tl k)) (scalarAdd c (toHeightmap tr k))) (beside (scalarAdd c (toHeightmap bl k)) (scalarAdd c (toHeightmap br k)))
	where
	c = h * e
	k = e * 0.9

heightmapToStr :: Heightmap -> String
heightmapToStr h = concat (map (foldr ((++) . ("\t" ++) . show) "\n") h)

displayHeightmap :: Heightmap -> IO()
displayHeightmap = (putStr . heightmapToStr)

onTop :: [[a]] -> [[a]] -> [[a]]
onTop [] xs = xs
onTop xs [] = xs
onTop xs ys = xs ++ ys

beside :: [[a]] -> [[a]] -> [[a]]
beside [] xs = xs
beside xs [] = xs
beside xs ys = [ x ++ y | (x, y) <- (zip xs ys)]

scalarAdd :: Double -> Heightmap -> Heightmap
scalarAdd 0 h = h
scalarAdd c h = map (map (+ c)) h

scalarMultiply :: Double -> Heightmap -> Heightmap
scalarMultiply k h = map (map (* k)) h

absHeightmap :: Heightmap -> Heightmap
absHeightmap h = map (map abs) h

smoothMap :: Heightmap -> Heightmap
smoothMap h = (transpose . (map smoothRow) . transpose . (map smoothRow)) h

smoothRow :: [Double] -> [Double]
smoothRow (x:y:z:xs)	= x : smoothRow (((x + z)/2):z:xs)
smoothRow a				= a

generatePPM :: Heightmap -> String
generatePPM hm = "P3\n" ++ w ++ " " ++ h ++ "\n" ++ show colour ++ "\n" ++ (hmToRGB ph colour)
	where
	h		= (show . length) hm
	w		= (show . length . head) hm
	ph		= absHeightmap hm
	colour	= maxColour ph

maxColour :: Heightmap -> Int
maxColour h = (round . maximum . concat) h

hmToRGB :: Heightmap -> Int -> String
hmToRGB h c = concat (map (toRGB c) (concat h))

toRGB :: Int -> Double -> String
toRGB c x	| x > 0 = strX ++ " " ++ strX ++ " " ++ strX ++ "\n"
			| otherwise = show c ++ " 0 0\n"
			where
			strX = (show . round) x

writePPM :: String -> String -> IO()
writePPM filename ppm = do
			outHandle <- openFile filename WriteMode
			hPutStr outHandle ppm
			hClose outHandle

toPositionmap :: Heightmap -> Positionmap
toPositionmap h = [ toPositionrow r y | (r, y) <- (zip h [0..])]

toPositionrow :: [Double] -> Double -> [(Double, Double, Double)]
toPositionrow zs y = [ (x, y, z) | (x, z) <- (zip [0..] zs) ]

multiplyVector :: Matrix -> Vector -> Vector
multiplyVector m v = toVector (map ((dotProduct v) . toVector) m)

dotProduct :: Vector -> Vector -> Double
dotProduct (x1, y1, z1) (x2, y2, z2) = x1 * x2 + y1 * y2 + z1 * z2

toVector :: [Double] -> Vector
toVector [x, y, z] = (x, y, z)

rotateX :: Double -> Vector -> Vector
rotateX a v = multiplyVector m v
	where
	m = [[1.0, 0.0, 0.0], [0.0, cos a, -(sin a)], [0.0, sin a, cos a]]

rotateY :: Double -> Vector -> Vector
rotateY a v = multiplyVector m v
	where
	m = [[cos a, 0.0, sin a], [0.0, 1.0, 0.0], [-(sin a), 0.0, cos a]]

rotateZ :: Double -> Vector -> Vector
rotateZ a v = multiplyVector m v
	where
	m = [[cos a, -(sin a), 0.0], [sin a, cos a, 0.0], [0.0, 0.0, 1.0]]

rotateV :: (Double, Double, Double) -> Vector -> Vector
rotateV (x, y, z) v = (rotateX x . rotateY y . rotateZ z) v

rotatePM :: (Double, Double, Double) -> Positionmap -> Positionmap
rotatePM r pm = map (map (rotateV r)) pm

rotateVS :: (Double, Double, Double) -> [(Vector, Vector)] -> [(Vector, Vector)]
rotateVS r vs = map (\(a, b) -> ((rotateV r a), (rotateV r b))) vs

rotateFS :: (Double, Double, Double) -> [Face] -> [Face]
rotateFS r fs = map (\((a, b, c), h) -> ((rotateV r a, rotateV r b, rotateV r c), h)) fs

scaleV :: Double -> Vector -> Vector
scaleV k (x, y, z) = ((k * x), (k * y), (k * z))

scalePM :: Double -> Positionmap -> Positionmap
scalePM k pm = map (map (scaleV k)) pm

scaleVS :: Double -> [(Vector, Vector)] -> [(Vector, Vector)]
scaleVS k vs = map (\(a, b) -> (scaleV k a, scaleV k b)) vs

scaleFS :: Double -> [Face] -> [Face]
scaleFS k fs = map (\((a, b, c), h) -> ((scaleV k a, scaleV k b, scaleV k c), h)) fs

rowSort :: (Ord a) => (a, a, a) -> (a, a, a) -> Ordering
rowSort (x1, y1, _) (x2, y2, _) | y1 > y2	= LT
								| y1 < y2	= GT
								| x1 > x2	= LT
								| x1 < x2	= GT
								| otherwise	= EQ

absRowSort :: (Ord a) => (a, a) -> (a, a) -> Ordering
absRowSort (x1, y1) (x2, y2)	| y1 > y2	= GT
								| y1 < y2	= LT
								| x1 > x2	= GT
								| x2 < x1	= LT
								| otherwise	= EQ

rfpixelSort :: (Ord a) => ((a, a, a), b) -> ((a, a, a), b) -> Ordering
rfpixelSort ((x1, y1, z1), _) ((x2, y2, z2), _)	| y1 > y2	= GT
										| y1 < y2	= LT
										| x1 > x2	= GT
										| x1 < x2	= LT
										| z1 > z2	= LT
										| z1 < z2	= GT
										| otherwise	= EQ

getVectors :: Positionmap -> [(Vector, Vector)]
getVectors pm = (concat . map pairRow) pm ++ (concat . map pairRow . transpose) pm

pairRow :: [Vector] -> [(Vector, Vector)]
pairRow []	= []
pairRow [x]	= []
pairRow (x:y:xs) = (x, y) : pairRow (y:xs)

getPixels :: [(RVector, RVector)] -> [RVector]
getPixels rvs = (nub . concat . map linePixels) rvs

linePixels :: (RVector, RVector) -> [RVector]
linePixels ((x1, y1), (x2, y2)) = walkLine ((x1, y1), (x2, y2)) deltaX deltaY devX devY
	where
	deltaX	= abs (x2 - x1)
	deltaY	= abs (y2 - y1)
	devX	= 2 * deltaX - deltaY
	devY	= 2 * deltaY - deltaX

walkLine :: (RVector, RVector) -> Int -> Int -> Int -> Int -> [RVector]
walkLine ((x1, y1), (x2, y2)) dx dy dex dey	| (x1, y1) == (x2, y2) = [(x2, y2)]
											| x1 == x2	= zip (repeat x1) (if y1 > y2 then [y2..y1] else [y1..y2])
											| y1 == y2	= zip (if x1 > x2 then [x2..x1] else [x1..x2]) (repeat y1)
											| otherwise = (x1, y1) : walkLine ((nx, ny), (x2, y2)) dx dy ndex ndey
	where
	nx		= if dex > 0 then (if x1 > x2 then x1 - 1 else x1 + 1) else x1
	ny		= if dey > 0 then (if y1 > y2 then y1 - 1 else y1 + 1) else y1
	ndex	= (if dex > 0 then dex - dy else dex) + dx
	ndey	= (if dey > 0 then dey - dx else dey) + dy

vsToRVs :: [(Vector, Vector)] -> [(RVector, RVector)]
vsToRVs vs = map (\((x1, y1, z1), (x2, y2, z2)) -> ((round x1, round y1), (round x2, round y2))) vs

pixelsToBM :: [RVector] -> Boolmap
pixelsToBM rvs = [ map (contains rvs) (zip [lb..rb] (repeat y)) | y <- [tb..bb]]
	where
	xs	= map (\(x, y) -> x) rvs
	ys	= map (\(x, y) -> y) rvs
	lb	= minimum xs
	rb	= maximum xs
	w	= rb - lb
	tb	= minimum ys
	bb	= maximum ys
	h	= bb - tb

contains :: (Eq a) => [a] -> a -> Bool
contains [] _ = False
contains (x:xs) t	| x == t	= True
					| otherwise	= contains xs t

bmToPPM :: Boolmap -> String
bmToPPM bm = "P3\n" ++ w ++ " " ++ h ++ "\n1\n" ++ (concat . map (\x -> if x then "1 1 1\n" else "0 0 0\n") . concat) bm
	where
	w = (show . length . head) bm
	h = (show . length) bm

pixelsToPPM :: [RVector] -> String
pixelsToPPM rvs = "P3\n" ++ show w ++ " " ++ show h ++ "\n1\n" ++ (getPixelsPPM lb rb tb bb lb tb . sortBy absRowSort) rvs
	where
	xs	= map (\(x, y) -> x) rvs
	ys	= map (\(x, y) -> y) rvs
	lb	= minimum xs
	rb	= maximum xs
	w	= rb - lb + 1
	tb	= minimum ys
	bb	= maximum ys
	h	= bb - tb + 1

getPixelsPPM :: Int -> Int -> Int -> Int -> Int -> Int -> [RVector] -> String
getPixelsPPM lb rb tb bb cx cy []			| cx > rb	= []
											| otherwise	= "0 0 0\n" ++ getPixelsPPM lb rb tb bb (cx + 1) cy []
getPixelsPPM lb rb tb bb cx cy ((x, y):rvs)	| cx > rb	= getPixelsPPM lb rb tb bb lb (cy + 1) ((x, y):rvs)
											| cx == x	= "1 1 1\n" ++ getPixelsPPM lb rb tb bb (cx + 1) cy rvs
											| cx /= x	= "0 0 0\n" ++ getPixelsPPM lb rb tb bb (cx + 1) cy ((x, y):rvs)
											| otherwise	= "ERROR: Out of bounds!"

getFaces :: Positionmap -> [Face]
getFaces []			= []
getFaces [r]		= []
getFaces (r1:r2:pm)	= pairFaces (zip r1 r2) ++ getFaces (r2:pm)

pairFaces :: [(Vector, Vector)] -> [Face]
pairFaces []					= []
pairFaces [a]					= []
pairFaces ((a, b):(c, d):fs)	= ((a, b, c), getMaxZ (a, c, b)) : ((b, c, d), getMaxZ (b, c, d)) : pairFaces ((c, d):fs)

getNormal :: (RFVector, RFVector, RFVector) -> Vector
getNormal (a, b, c) = crossProduct (pointsToVector (b, a)) (pointsToVector (b, c))

pointsToVector :: (RFVector, RFVector) -> Vector
pointsToVector ((x1, y1, z1), (x2, y2, z2)) = (fromIntegral (x2 - x1), fromIntegral (y2 - y1), fromIntegral (z2 - z1))

crossProduct :: Vector -> Vector -> Vector
crossProduct (x1, y1, z1) (x2, y2, z2) = ((y1 * z2 - z1 * y2), (z1 * x2 - z2 * x1), (x1 * y2 - x2 * y1))

getMaxZ :: (Vector, Vector, Vector) -> Double
getMaxZ ((_, _, z1), (_, _, z2), (_, _, z3)) = max z1 (max z2 z3)

fsToRFs :: [Face] -> [RFace]
fsToRFs = map fToRF

fToRF :: Face -> RFace
fToRF ((a, b, c), h) = ((vToRFV a, vToRFV b, vToRFV c), h)

vToRFV :: Vector -> RFVector
vToRFV (x, y, z) = (round x, round y, round z)

rfacesToFPixels :: Vector -> [RFace] -> [RFPixel]
rfacesToFPixels sv fs = (prioritiseRFPixels . concat . map getFacePixels) fs

prioritiseRFPixels :: [((Int, Int, Int), RGB)] -> [RFPixel]
prioritiseRFPixels ps = (map (\((x, y, z), c) -> ((x, y), c)) . nubBy (\((x1, y1, _), _) ((x2, y2, _), _) -> (x1, y1) == (x2, y2)) . sortBy rfpixelSort) ps

getFacePixels :: RFace -> [((Int, Int, Int), RGB)]
getFacePixels ((a, b, c), h) = (map (\(x, y) -> ((x, y, fz), colour)) . fillShape . sortBy absRowSort) (linePixels (ra, rb) ++ linePixels (rb, rc) ++ linePixels (rc, ra))
	where
	ra	= (\(x, y, z) -> (x, y)) a
	rb	= (\(x, y, z) -> (x, y)) b
	rc	= (\(x, y, z) -> (x, y)) c
	n	= normaliseVector (getNormal (a, b, c))
	red		= div (abs (floor ((dotProduct n (normaliseVector (0,0,1))) * 65535))) 4 + 32000
	green	= div (abs (floor ((dotProduct n (normaliseVector (0,1,0))) * 65535))) 4 + 32000
	blue	= div (abs (floor ((dotProduct n (normaliseVector (1,0,0))) * 65535))) 4 + 32000
	az  = (\(_, _, z) -> z) a
	bz  = (\(_, _, z) -> z) b
	cz  = (\(_, _, z) -> z) c
	fz  = div (az + bz + cz) 3
	colour = (abs red, abs green, abs blue)


--maybe not fine
fillShape :: [RVector] -> [RVector]
fillShape []	= []
fillShape [x]	= [x]
fillShape ((x1, y1):(x2, y2):xs)	| y1 == y2	= (zip [x1..x2] (repeat y1)) ++ fillShape ((x2, y2):xs)
									| otherwise	= (x1, y1) : fillShape ((x2, y2):xs)
--Fine
normaliseVector :: Vector -> Vector
normaliseVector (x, y, z) = (k * x, k * y, k * z)
	where
	k = 1 / (sqrt (x ^ 2 + y ^ 2 + z ^ 2))

rfpixelsToPPM :: [RFPixel] -> String
rfpixelsToPPM ps = "P3\n" ++ show w ++ " " ++ show h ++ "\n65535\n" ++ getRFPixelsPPM lb rb tb bb lb tb ps
	where
	xs	= map (\((x, _), _) -> x) ps
	ys	= map (\((_, y), _) -> y) ps
	lb	= minimum xs
	rb	= maximum xs
	w	= rb - lb + 1
	tb	= minimum ys
	bb	= maximum ys
	h	= bb - tb + 1

getRFPixelsPPM :: Int -> Int -> Int -> Int -> Int -> Int -> [RFPixel] -> String
getRFPixelsPPM lb rb tb bb cx cy []							| cx > rb	= []
															| otherwise	= "0 0 0\n" ++ getRFPixelsPPM lb rb tb bb (cx + 1) cy []
getRFPixelsPPM lb rb tb bb cx cy (((x, y), (r, g, b)):rvs)	| cx > rb	= getRFPixelsPPM lb rb tb bb lb (cy + 1) (((x, y), (r, g, b)):rvs)
															| cx == x	= (show r ++ " " ++ show g ++ " " ++ show b ++ "\n") ++ getRFPixelsPPM lb rb tb bb (cx + 1) cy rvs
															| cx /= x	=  "0 0 0\n" ++ getRFPixelsPPM lb rb tb bb (cx + 1) cy (((x, y), (r, g, b)):rvs)
															| otherwise	= "ERROR: Out of bounds!"
